/* ******************************************************************** *

 INFORMER 1.1 - Copyright (c) 1992-1994, Bob Flanders and Michael Holmes
  Version 1.0 published in PC Magazine C Lab notes, ZD Press, 1992
  Version 1.1 published in PC Magazine, January 25, 1994

* ******************************************************************** *

   Compile INFORMER using the following vendor specific commands.

      Borland C++ version 3.x
        BCC -O2 -mt informer.c

      Microsoft C/C++ version 7.0
        CL /AS informer.c

 * ******************************************************************** */


#include <stdio.h>                          /* standard i/o library     */
#include <stdlib.h>                         /* ANSI compatibility       */
#include <dos.h>                            /* DOS rtn definitions      */
#include <string.h>                         /* string functions         */
#include <stdarg.h>                         /* argument functions       */
#include <process.h>                        /* exec functions           */
#pragma  pack(1)                            /* pack to byte alignment   */


#define NOT !                               /* logical not              */
#define DATE(x) x.month, x.day, x.year                  /* date fields  */
#define TIME(x) x.hour, x.minute, x.second, x.hsecond   /* time fields  */
#define UINT  unsigned int                  /* unsigned integer type    */


#if defined(__TURBOC__) || defined(__BORLANDC__)    /* BC++ 3.x         */
#define  _BC_VER                            /* Borland shorthand        */
#define INT_PARMS UINT bp, UINT di,         /* interrupt calling conv  */\
            UINT si, UINT ds, UINT es, UINT dx, UINT cx,                 \
            UINT bx, UINT ax, UINT cs, UINT ip, UINT flags
#endif


#if defined(_MSC_VER)                       /* MSC C/C++ 7              */
#define MK_FP(s,o)                                                       \
          (void far *)                                                   \
              (((long) s << 16) | (long) o) /* make a far pointer       */
#define INT_PARMS UINT es, UINT ds,         /* interrupt calling conv  */\
            UINT di, UINT si, UINT bp, UINT sp, UINT bx,                 \
            UINT dx, UINT cx, UINT ax, UINT ip, UINT cs, UINT flags
#endif


/* ******************************************************************** *
 *      routine definitions                                             *
 * ******************************************************************** */

void    interrupt far int08_new(INT_PARMS), /* timer interrupt          */
        interrupt far int10_new(INT_PARMS), /* video interrupt          */
        interrupt far int14_new(INT_PARMS), /* serial interrupt         */
        interrupt far int16_new(INT_PARMS), /* keyboard                 */
        interrupt far int17_new(INT_PARMS), /* printer interrupt        */
        interrupt far int21_new(INT_PARMS), /* DOS interrupt            */
        interrupt far int25_new(INT_PARMS), /* Absolute disk read       */
        interrupt far int26_new(INT_PARMS), /* Absolute disk write      */
        interrupt far int67_new(INT_PARMS); /* EMS interrupt            */


/* ******************************************************************** *
 *      globals                                                         *
 * ******************************************************************** */

int     rc = 1,                             /* errorlevel return code   */
        pgm_token = 1,                      /* program name token       */
        enabled,                            /* flag for new int handlers*/
        tsr,                                /* monitored pgm went tsr   */
        ints_changed;                       /* interrupt vectors changed*/

char   *log_name,                           /* log file name            */
       *cmd[3];                             /* command stack            */

long    int21_cnt[256];                     /* DOS subfunction counts   */

struct  dosdate_t start_d, end_d;           /* date fields              */
struct  dostime_t start_t, end_t;           /* ..and time fields        */


struct                                      /* usage by classification  */
    {
    long int_tcnt;                              /* count for this type  */
    char *int_tdes;                             /* description          */
    } int_type[] =
        { { 0, "Console Calls" },
          { 0, "Disk Calls" },
          { 0, "Printer Calls" },
          { 0, "Video Calls" },
          { 0, "Memory Management Calls" },
          { 0, "Serial Communications" },
          { 0, "Operating System Calls" }
        };

#define INT_NONE -1
#define INT_CON   0
#define INT_DISK  1
#define INT_PRT   2
#define INT_VID   3
#define INT_MEM   4
#define INT_COMM  5
#define INT_DOS   6
#define INT_TYPE sizeof(int_type) / sizeof(int_type[0])


struct                                      /* i/o byte counts          */
    {
    char *io_desc;                              /* description          */
    long io_read,                               /* bytes read           */
         io_write;                              /* bytes written        */
    } io_cnt[] =
        { { "Keyboard" },
          { "Video" },
          { "Serial" },
          { "Printer" },
          { "File" }
        };

#define IO_KB     0
#define IO_VID    1
#define IO_COMM   2
#define IO_PRT    3
#define IO_FILE   4
#define IO_CNT    sizeof(io_cnt) / sizeof(io_cnt[0])


struct                                      /* interrupt list           */
    {
    int  int_nbr;                               /* interrupt number     */
    char int_itype,                             /* interrupt type       */
        *int_text;                              /* description          */
    void (interrupt far *int_new)(),            /* new routine address  */
         (interrupt far *int_old)(void);        /* old int address      */
    long int_cnt;                               /* int usage count      */
    } int_list[] =
        { { 0x08, INT_NONE, "Timer",      int08_new },
          { 0x10, INT_VID,  "Video",      int10_new },
          { 0x14, INT_COMM, "Serial I/O", int14_new },
          { 0x16, INT_CON,  "Keyboard",   int16_new },
          { 0x17, INT_PRT,  "Printer",    int17_new },
          { 0x67, INT_MEM,  "EMS",        int67_new },
          { 0x21, INT_NONE, "DOS",        int21_new },
          { 0x25, INT_DISK, "Disk read",  int25_new },
          { 0x26, INT_DISK, "Disk write", int26_new }
        };

#define INT_08    0                         /* int_list entries         */
#define INT_10    1
#define INT_14    2
#define INT_16    3
#define INT_17    4
#define INT_67    5
#define INT_21    6
#define INT_25    7
#define INT_26    8
#define INT_LIST  sizeof(int_list) / sizeof(int_list[0])


struct                                      /* int 21 subfunctions      */
    {
    char int21_type;                            /* type of function     */
    char *int21_text;                           /* function description */
    } int21_list[] =
        { { INT_DOS,  "Program termination" },
          { INT_CON,  "Keyboard input" },
          { INT_VID,  "Display output" },
          { INT_CON,  "AUX input" },
          { INT_CON,  "AUX output" },
          { INT_PRT,  "Printer output" },
          { INT_CON,  "Direct console I/O" },
          { INT_CON,  "Direct STDIN input, no echo" },
          { INT_CON,  "Keyboard input, no echo" },
          { INT_VID,  "Print string" },
          { INT_CON,  "Buffered keyboard input" },
          { INT_CON,  "Check standard input status" },
          { INT_CON,  "Clear keyboard buffer" },
          { INT_DISK, "Disk reset" },
          { INT_DISK, "Select disk" },
          { INT_DISK, "Open disk file" },
          { INT_DISK, "Close disk file" },
          { INT_DISK, "Search first using FCB" },
          { INT_DISK, "Search next using FCB" },
          { INT_DISK, "Delete file via FCB" },
          { INT_DISK, "Sequential disk file read" },
          { INT_DISK, "Sequential disk record write" },
          { INT_DISK, "Create a disk file" },
          { INT_DISK, "Rename file via FCB" },
          { INT_DOS,  "Internal Use" },
          { INT_DOS,  "Get default disk number" },
          { INT_DOS,  "Set disk transfer area" },
          { INT_DISK, "Get allocation for default drive" },
          { INT_DISK, "Get allocation for specific drive" },
          { INT_DOS,  "Internal Use" },
          { INT_DOS,  "Internal Use" },
          { INT_DOS,  "Get default drive parm block" },
          { INT_DOS,  "Internal Use" },
          { INT_DISK, "Random disk record read" },
          { INT_DISK, "Random disk record write" },
          { INT_DISK, "Get file size" },
          { INT_DISK, "Set random record field" },
          { INT_DOS,  "Set interrupt vector" },
          { INT_MEM,  "Create PSP" },
          { INT_DISK, "Random block read" },
          { INT_DISK, "Random block write" },
          { INT_DOS,  "Parse filename" },
          { INT_DOS,  "Get date" },
          { INT_DOS,  "Set date" },
          { INT_DOS,  "Get time" },
          { INT_DOS,  "Set time" },
          { INT_DOS,  "Set verify flag" },
          { INT_DOS,  "Get disk transfer area" },
          { INT_DOS,  "Get DOS version" },
          { INT_DOS,  "Terminate and stay resident" },
          { INT_DOS,  "Get drive parameter block" },
          { INT_DOS,  "Extended ^Break/Get boot drive" },
          { INT_DOS,  "Get CritSectFlag pointer" },
          { INT_DOS,  "Get interrupt vector" },
          { INT_DOS,  "Get disk space" },
          { INT_DOS,  "Get/set SWITCHAR" },
          { INT_DOS,  "Get/set country information" },
          { INT_DISK, "Create a subdirectory" },
          { INT_DISK, "Remove a directory entry" },
          { INT_DISK, "Change current directory" },
          { INT_DISK, "Create a file with handle" },
          { INT_DISK, "Open disk file with handle" },
          { INT_DISK, "Close a file with handle" },
          { INT_DISK, "Read from file with handle" },
          { INT_DISK, "Write to file with handle" },
          { INT_DISK, "Delete a file" },
          { INT_DISK, "Move file pointer" },
          { INT_DISK, "Get/put file attributes" },
          { INT_DISK, "IOCTL" },
          { INT_DISK, "Create duplicate handle" },
          { INT_DISK, "Force duplicate handle" },
          { INT_DOS,  "Get current directory" },
          { INT_MEM,  "Allocate memory" },
          { INT_MEM,  "Free memory" },
          { INT_MEM,  "Adjust memory block size" },
          { INT_DISK, "Load or execute" },
          { INT_DOS,  "Quit with exit code" },
          { INT_DOS,  "Get exit code of subprogram" },
          { INT_DISK, "Find first ASCIIZ" },
          { INT_DISK, "Find next ASCIIZ" },
          { INT_DOS,  "Set PSP segment" },
          { INT_DOS,  "Get PSP segment" },
          { INT_DOS,  "Get list of lists" },
          { INT_DOS,  "Translate BIOS parameter block" },
          { INT_DOS,  "Get verify flag" },
          { INT_MEM,  "Create PSP" },
          { INT_DISK, "Rename a file" },
          { INT_DISK, "Get/set file's date/time" },
          { INT_MEM,  "Get/set memory allocation strategy" },
          { INT_DOS,  "Get extended error code" },
          { INT_DISK, "Create unique file" },
          { INT_DISK, "Create new file" },
          { INT_DISK, "Lock/unlock file access" },
          { INT_NONE, "Internal Use" },
          { INT_DOS,  "DOS 3.1 + Microsoft Networks" },
          { INT_DOS,  "Network" },
          { INT_DOS,  "Resolve path string" },
          { INT_NONE, "Unused" },
          { INT_DOS,  "Get PSP address" },
          { INT_DOS,  "2-byte character support" },
          { INT_NONE, "Unused" },
          { INT_DOS,  "Country-dependent support" },
          { INT_DOS,  "Get/set global code page table" },
          { INT_DOS,  "Set handle count" },
          { INT_DISK, "Commit file" },
          { INT_DISK, "Get/set disk serial number" },
          { INT_NONE, "Unused" },
          { INT_NONE, "Unused" },
          { INT_DISK, "Extended open/create" }
        };

#define INT21_LIST sizeof(int21_list) / sizeof(int21_list[0])


struct  EMS_hdr                             /* fake EMS driver header   */
    {
    long ems_next;                              /* next pointer         */
    int  ems_attr,                              /* attributes           */
         ems_strategy,                          /* strategy offset      */
         ems_int;                               /* interrupt offset     */
    char ems_name[9],                           /* name field, EMMXXXX0 */
         ems_jump;                              /* far jump instruction */
    void (interrupt far *ems_interrupt)();      /* new routine address  */
    char ems_filler[16];                        /* filler paragraph     */
    };


/* ******************************************************************** *
 *      messages and strings                                            *
 * ******************************************************************** */

char    copyright[]     = "INFORMER 1.1  Copyright (c) 1992-1994, "
                          "Bob Flanders and Michael Holmes\n"
                          "Version 1.0 published in PC Magazine C Lab "
                          "Notes, ZD Press, 1992\n"
                          "Version 1.1 published in PC Magazine, "
                          "January 25, 1994\n\n",
        stop_here[]     = "\nStopping at user's request\n",
        not_found[]     = "Requested program not found\n",
        bad_log[]       = "Could not open output log file %s\n",
        header[]        = "\n--------------- Informer Statistics "
                          "---------------\n",
        cmd_msg[]       = "    Command Line: %s %s",
        time_msg[]      = "\n    %s: %2d:%02d:%02d",
        date_msg[]      = " on %02d/%02d/%d",
        starting[]      = "  Started at",
        ending[]        = "Completed at",
        run_time[]      = "\n    Elapsed Time: %2d:%02d:%02d.%02d",
        term_msg[]      = "\nTermination Code: %d\n",
        int_hdr[]       = "\nInterrupt Usage\n",
        int_msg[]       = "%8ld Int %02X - %s interrupt\n",
        int21_hdr[]     = "\nInt 21 (DOS) Breakdown by Function\n",
        int21_msg[]     = "%8ld Fnc %02X - %s\n",
        type_hdr[]      = "\nUsage by Classification\n",
        type_msg[]      = "%8ld - %s\n",
        io_hdr[]        = "\nInput/Output Byte Counts\n"
                          "    read  written\n",
       *io_msg[2]     = { "%8ld %8s - %s\n",
                          "%8ld %8ld - %s\n" },
        tsr_msg[]       = "Monitored program terminated and "
                          "stayed resident.\n",
        ints_msg[]      = "Monitored program changed interrupt vectors.\n",
        reboot_msg[]    = "Reboot suggested.\n",
        null_line[]     = "",
        help[]          =
            "  Usage:  INFORMER  [/output]  program  [options]\n\n"
            "Options:  output  specifies the filename where INFORMER "
            "writes its report\n"
            "          program is the program INFORMER runs and "
            "monitors\n"
            "          options are the command line arguments for the "
            "monitored program\n";



/* ******************************************************************** *
 *      quit_with() -- give an error message, then return to DOS        *
 * ******************************************************************** */

void    quit_with(char *msg, ...)           /* string to print          */
{
va_list list;                               /* variable list            */


va_start(list, msg);                        /* set up variable list     */
vprintf(msg, list);                         /* give error message ..    */
exit(rc);                                   /* ..and then quit          */

}



/* ******************************************************************** *
 *      check_arguments() -- check command line arguments               *
 * ******************************************************************** */

void    check_arguments(int  ac,            /* DOS cmd line token count */
                       char *av[])          /* ..token strings          */
{
int     i;                                  /* loop control             */
char    *p, *q,                             /* string pointers          */
        cmd_buff[128];                      /* ..and work buffer        */
FILE    *log;                               /* output log file handle   */


if (NOT strcmp(av[1], "/?"))                /* q. user needs help?      */
    quit_with(help);                        /* a. yes .. give help/quit */

if (av[1][0] == '/')                        /* q. user wants a log file?*/
    {
    log_name = &av[1][1];                   /* a. yes .. save name      */

    if (NOT (log = fopen(log_name, "w")))   /* q. log open sucessfully? */
        quit_with(bad_log, log_name);       /* a. no .. quit w/error msg*/

    fclose(log);                            /* close log file           */
    i = 2;                                  /* ..and skip to next token */
    }
 else
    i = 1;                                  /* else .. start w/first one*/

if (ac <= i)                                /* q. missing program name? */
    quit_with(help);                        /* a. yes .. give help/quit */

memset(cmd_buff, 0, sizeof(cmd_buff));      /* clear buffer then ..     */
_fstrncpy(cmd_buff, MK_FP(_psp, 0x81),      /* copy command line locally*/
        * (char far *) MK_FP(_psp, 0x80));

if (i == 2)                                 /* q. log name given?       */
    p = strstr(cmd_buff, log_name) +        /* a. yes .. skip past      */
            strlen(log_name);               /* ..the log file name      */
 else
    p = &cmd_buff[1];                       /* else .. start at start   */

for (; *p; p++)                             /* find start of command    */
    if (*p != ' ')                          /* q. find command name?    */
        break;                              /* a. yes .. exit loop      */

if ((i = strcspn(p, " /")) != strlen(p))    /* q. find end of cmd name? */
    {
    q = p + i;                              /* a. yes .. save pointer   */
    cmd[1] = strdup(q);                     /* ..and save a copy        */

    *q = 0;                                 /* terminate cmd name       */
    }
 else
    cmd[1] = "";                            /* else .. use null string  */

cmd[0] = strdup(p);                         /* save a copy of cmd name  */

}



/* ******************************************************************** *
 *      int08_new() -- timer interrupt appendage                        *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma option -Od                  /* no optimization                  */
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int08_new(INT_PARMS)
{


if (enabled)                                /* q. collecting data?      */
    int_list[INT_08].int_cnt++;             /* a. yes .. tally ticks    */

_chain_intr(*int_list[INT_08].int_old);     /* jump to old int handler  */

}



/* ******************************************************************** *
 *      int10_new() -- video interrupt                                  *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int10_new(INT_PARMS)
{


switch (ax >> 8)                            /* based on function code   */
    {
                                            /* read functions           */
    case 0x08:                              /* character & attributes   */
    case 0x0d:                              /* pixel                    */
        io_cnt[IO_VID].io_read++;           /* bump read byte count     */
        break;                              /* ..and continue           */

                                            /* write functions          */
    case 0x09:                              /* character & attribute    */
    case 0x0a:                              /* character only           */
    case 0x0c:                              /* pixel                    */
    case 0x0e:                              /* tty mode                 */
        io_cnt[IO_VID].io_write++;          /* bump write byte count    */
        break;                              /* ..and continue           */

    case 0x13:                              /* write string             */
        io_cnt[IO_VID].io_write += cx;      /* add characters written   */
    }

int_list[INT_10].int_cnt++;                 /* tally video int requests */
_chain_intr(*int_list[INT_10].int_old);     /* jump to old int handler  */

}



/* ******************************************************************** *
 *      int14_new() -- serial i/o                                       *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int14_new(INT_PARMS)
{


switch (ax >> 8)                            /* based on function code   */
    {
    case 0x01:                              /* write character          */
        io_cnt[IO_COMM].io_write++;         /* bump write byte count    */
        break;                              /* ..and continue           */

    case 0x02:                              /* read character           */
        io_cnt[IO_COMM].io_read++;          /* bump read byte count     */
        break;                              /* ..and continue           */
    }

int_list[INT_14].int_cnt++;                 /* tally by serial i/o ints */
_chain_intr(*int_list[INT_14].int_old);     /* jump to old int handler  */

}



/* ******************************************************************** *
 *      int16_new() -- keyboard interrupt                               *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int16_new(INT_PARMS)
{


if (NOT (ax & 0xef00))                      /* q. getting a character?  */
    io_cnt[IO_KB].io_read++;                /* a. yes .. cnt keystrokes */

int_list[INT_16].int_cnt++;                 /* tally by keyboard calls  */
_chain_intr(*int_list[INT_16].int_old);     /* jump to old int handler  */

}



/* ******************************************************************** *
 *      int17_new() -- printer interrupts                               *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int17_new(INT_PARMS)
{


if (NOT (ax & 0xff00))                      /* q. printing a character? */
    io_cnt[IO_PRT].io_write++;              /* a. yes .. cnt each one   */

int_list[INT_17].int_cnt++;                 /* tally printer interrupts */
_chain_intr(*int_list[INT_17].int_old);     /* jump to old int handler  */

}



/* ******************************************************************** *
 *      int21_handler() -- interrupt 21 handler                         *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int21_new(INT_PARMS)
{


if (enabled)                                /* q. collecting data?      */
    {
    int_list[INT_21].int_cnt++;             /* a. yes ..count all ints  */
    int21_cnt[ax >> 8]++;                   /* ..and tally by subfunc   */

    switch (ax >> 8)                        /* based on function code   */
        {
        case 0x3f:                          /* read from file/device    */
            io_cnt[IO_FILE].io_read += cx;  /* count bytes read         */
            break;                          /* ..and continue           */

        case 0x40:                          /* write to file/device     */
            io_cnt[IO_FILE].io_write += cx; /* count bytes written      */
            break;                          /* ..and continue           */

        case 0x31:                          /* terminate & stay resident*/
            tsr = 1;                        /* ..set flag for later     */

        case 0:                             /* terminate                */
        case 0x4c:                          /* terminate w/return code  */
            enabled = 0;                    /* quit collecting stats    */
            break;                          /* ..and continue           */
        }
    }

 else if (ax == 0x4b00)                     /* q. load/execute call?    */
    enabled = 1;                            /* a. yes ..start collecting*/

_chain_intr(*int_list[INT_21].int_old);     /* then jump to old int21   */

}



/* ******************************************************************** *
 *      int25_handler() -- absolute disk reads                          *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int25_new(INT_PARMS)
{


int_list[INT_25].int_cnt++;                 /* count disk reads         */
_chain_intr(*int_list[INT_25].int_old);     /* then jump to old int25   */

}



/* ******************************************************************** *
 *      int26_handler() -- absolute disk writes                         *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int26_new(INT_PARMS)
{


int_list[INT_26].int_cnt++;                 /* count disk writes        */
_chain_intr(*int_list[INT_26].int_old);     /* then jump to old int26   */

}



/* ******************************************************************** *
 *      int67_handler() -- EMS interrupt handler                        *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma argsused                    /* hold unused argument messages    */
#endif

void    interrupt far int67_new(INT_PARMS)
{


if (enabled)                                /* q. collecting data?      */
    int_list[INT_67].int_cnt++;             /* a. yes .. count EMS call */

_chain_intr(*int_list[INT_67].int_old);     /* then jump to old int67   */

}



/* ******************************************************************** *
 *      install_handlers() -- install interrupt handlers                *
 * ******************************************************************** */

#ifdef _BC_VER
#pragma option -O                   /* optimization re-enabled          */
#endif

void    install_handlers(void)
{
int     i;                                  /* loop control             */
struct  EMS_hdr huge *ep;                   /* EMS header pointer       */
union
    {
    char    huge *p;                        /* work pointer             */
    long    q;                              /* work address             */
    } x;


x.p = (char huge *)                         /* get enough memory        */
        malloc(sizeof(struct EMS_hdr));     /* ..for the EMS header     */
x.p += 15;                                  /* start at the beginning   */
x.q = (x.q & 0xffff0000L) +                 /* ..and adjust huge pointer*/
        ((x.q & 0xfff0) << 12);             /*   (covers MSC C7 bug)    */

ep = (struct EMS_hdr huge *) x.p;           /* set up EMS hdr pointer   */
ep->ems_next = -1;                          /* initialize EMS header    */
ep->ems_attr = 0xc000;                      /* ..with attributes        */
ep->ems_strategy = FP_OFF(ep->ems_jump);    /* ..offset to strategy     */
ep->ems_int = FP_OFF(ep->ems_jump);         /* ..and interrupt routine  */
_fmemcpy(ep->ems_name, "EMMXXXX0", 8);      /* ..device name            */
ep->ems_jump = 0xea;                        /* ..far jump               */
ep->ems_interrupt = int67_new;              /* ..our interrupt routine  */

int_list[INT_67].int_new =                  /* save the new EMS         */
       (void far *) &ep->ems_jump;          /* ..addr in out table      */

for (i = 0; i < INT_LIST; i++)              /* for each interrupt       */
    {
    int_list[i].int_old =                   /* get old interrupt ..     */
            (void far *) _dos_getvect(      /* ..vector address         */
            int_list[i].int_nbr);

    _dos_setvect(int_list[i].int_nbr,       /* install new ..           */
                    int_list[i].int_new);   /* ..interrupt handler      */
    }

_dos_getdate(&start_d);                     /* get start date..         */
_dos_gettime(&start_t);                     /* ..and start time         */

}



/* ******************************************************************** *
 *      remove_handlers() -- remove interrupt handlers                  *
 * ******************************************************************** */

void    remove_handlers(void)
{
int     i;                                  /* loop control             */


enabled = 0;                                /* stop collecting data     */

_dos_getdate(&end_d);                       /* get ending date..        */
_dos_gettime(&end_t);                       /* ..and time               */

for (i = 0; i < INT_LIST; i++)              /* for each interrupt       */
    {
    if (_dos_getvect(int_list[i].int_nbr)   /* q. our interrupt overlaid*/
                != int_list[i].int_new)     /* ..by the monitored pgm?  */
        ints_changed = 1;                   /* a. yes .. set flag       */

    _dos_setvect(int_list[i].int_nbr,       /* restore old ..           */
                    int_list[i].int_old);   /* ..interrupt handler      */
    }
}



/* ******************************************************************** *
 *      user_report() -- give user the final report                     *
 * ******************************************************************** */

void    user_report(void)
{
int     i,                                  /* loop control             */
        hdr,                                /* header flag              */
        hh, mm, ss;                         /* hours, minutes, seconds  */
long    hs;                                 /* ..and hundredths seconds */
FILE   *log;                                /* output log file handle   */

if (rc == -1)                               /* q. find requested pgm?   */
    quit_with(not_found);                   /* a. no .. quit w/error msg*/

if (log_name)                               /* q. log file specified?   */
    {                                       /* a. yes .. try opening it */
    if (NOT (log = fopen(log_name, "w")))   /* q. log open sucessfully? */
        quit_with(bad_log, log_name);       /* a. no .. quit w/error msg*/
    }
 else
    log = stdout;                           /* use console as log file  */

fprintf(log, header);                       /* give summary header      */
fprintf(log, cmd_msg, cmd[0], cmd[1]);      /* ..command line           */
fprintf(log, time_msg, starting,            /* ..start time             */
                TIME(start_t));
fprintf(log, date_msg, DATE(start_d));      /* ..and start date         */
fprintf(log, time_msg, ending, TIME(end_t));/* ..end'g time             */

if (strncmp((void *) &start_d,              /* q. date change while     */
                (void *) &end_d,            /* ..we were running the    */
                sizeof(struct dostime_t)))  /* ..user's program?        */
    fprintf(log, date_msg, DATE(end_d));    /* a. yes .. give end'g date*/

if ((hs = int_list[INT_08].int_cnt) == 0)   /* q. any ticks happen?     */
    hs = 1;                                 /* a. no .. set to one      */

hs *= 5.49254932;                           /* ticks to hundredth secs  */
hh = (int) (hs / 360000L);                  /* get hours                */
hs %= 360000L;                              /* ..then remove from total */
mm = (int) (hs / 6000);                     /* get minutes              */
hs %= 6000;                                 /* ..then remove from total */
ss = (int) (hs / 100);                      /* get seconds              */
hs %= 100;                                  /* ..leaving hundredths secs*/

fprintf(log, run_time, hh, mm, ss, hs);     /* give run-time message    */
fprintf(log, term_msg, rc);                 /* ..termination code       */

fprintf(log, int_hdr);                      /* give int header message  */

for (i = 1; i < INT_LIST; i++)              /* for each interrupt       */
    if (int_list[i].int_cnt)                /* q. any of that interrupt?*/
        {
        fprintf(log, int_msg,               /* a. yes .. display message*/
                int_list[i].int_cnt,        /* ..and count              */
                int_list[i].int_nbr,        /* ..and interrupt number   */
                int_list[i].int_text);      /* ..and interrupt name     */

        if (int_list[i].int_itype != INT_NONE)       /* q. classifiable?*/
            int_type[int_list[i].int_itype].int_tcnt /* a. yes .. build */
                += int_list[i].int_cnt;              /* ..counts by type*/
        }

for (i = hdr = 0; i < INT21_LIST; i++)      /* for each known function  */
    if (int21_cnt[i] > 0)                   /* q. function used?        */
        {                                   /* a. yes .. report on count*/
        if (NOT hdr)                        /* q. header given yet?     */
            {
            fprintf(log, int21_hdr);        /* a. no .. display header  */
            hdr++;                          /* ..and show header given  */
            }

        fprintf(log, int21_msg,             /* a. yes .. report count   */
                int21_cnt[i], i,
                int21_list[i].int21_text);

        if (int21_list[i].int21_type != INT_NONE)       /* q. collect?  */
            int_type[int21_list[i].int21_type].int_tcnt /* yes .. add   */
                    += int21_cnt[i];                    /* ..to totals  */
        }

fprintf(log, type_hdr);                     /* give usage hdr message   */
for (i = 0; i < INT_TYPE; i++)              /* print each type of call  */
    if (int_type[i].int_tcnt)               /* q. counter non-zero?     */
        fprintf(log, type_msg,              /* a. yes.. give usage stats*/
            int_type[i].int_tcnt,
            int_type[i].int_tdes);

for (i = hdr = 0; i < IO_CNT; i++)          /* print each byte cnt type */
    if (io_cnt[i].io_read ||                /* q. anything to report..  */
                    io_cnt[i].io_write)     /* ..inbound or outbound?   */
        {                                   /* a. yes .. print a line   */
        if (NOT hdr)                        /* q. header given yet?     */
            {
            fprintf(log, io_hdr);           /* a. no .. display header  */
            hdr++;                          /* ..and show header given  */
            }

        if (i == IO_KB)                     /* q. byte cnt for keyboard?*/
            fprintf(log, io_msg[0],         /* a. yes .. give read cnts */
                io_cnt[i].io_read, "",
                io_cnt[i].io_desc);
         else
            fprintf(log, io_msg[1],         /* else .. print read..     */
                io_cnt[i].io_read,          /* ..and write counts       */
                io_cnt[i].io_write,
                io_cnt[i].io_desc);
        }

if (tsr || ints_changed)                    /* q. need a line skipped?  */
    fprintf(log, "\n");                     /* a. yes .. set up for msgs*/

if (tsr)                                    /* q. program go tsr?       */
    fprintf(log, tsr_msg);                  /* a. yes .. give message   */

if (ints_changed)                           /* q. pgm take any ints?    */
    fprintf(log, ints_msg);                 /* a. yes .. give message   */

if (tsr || ints_changed)                    /* q. need rebooting?       */
    fprintf(log, reboot_msg);               /* a. yes .. give message   */

if (log_name)                               /* q. log file being used?  */
    fclose(log);                            /* a. yes .. close log file */

rc = 0;                                     /* reset return code        */
quit_with(null_line);                       /* ..and return to DOS      */

}



/* ******************************************************************** *
 *      main()                                                          *
 * ******************************************************************** */

void    main(int  ac,                       /* DOS cmd line token count */
             char *av[])                    /* ..token strings          */
{


printf(copyright);                          /* give copyright message   */

check_arguments(ac, av);                    /* check arguments          */
install_handlers();                         /* install interrupt rtns   */
rc = spawnvp(P_WAIT, cmd[0], &cmd[0]);      /* run requested program    */
remove_handlers();                          /* de-install interrupt rtns*/
user_report();                              /* give user feedback..     */
                                            /* ..and return to DOS      */
}



